@@ -298,9 +298,12 @@ prepareInArg arg = do
             (convert name $ hToF (argType arg) (transfer arg))
             (do
               let maybeName = "maybe" <> ucFirst name
+              nullPtr <- nullPtrForType (argType arg) >>= \case
+                Nothing -> terror $ "Unexpected non-pointer type " <> tshow (argType arg)
+                Just null -> pure null
               line $ maybeName <> " <- case " <> name <> " of"
               indent $ do
-                line $ "Nothing -> return nullPtr"
+                line $ "Nothing -> return " <> nullPtr
                 let jName = "j" <> ucFirst name
                 line $ "Just " <> jName <> " -> do"
                 indent $ do
@@ -360,7 +363,7 @@ prepareInCallback arg callback@(Callback {cbCallable = cb}) expose = do
               let maybeName = "maybe" <> ucFirst name
               line $ maybeName <> " <- case " <> name <> " of"
               indent $ do
-                line $ "Nothing -> return (castPtrToFunPtr nullPtr)"
+                line $ "Nothing -> return FP.nullFunPtr"
                 let jName = "j" <> ucFirst name
                     jName' = prime jName
                 line $ "Just " <> jName <> " -> do"
@@ -388,17 +391,24 @@ prepareInoutArg arg = do
   name' <- prepareInArg arg
   ft <- foreignType $ argType arg
   allocInfo <- typeAllocInfo (argType arg)
+
   case allocInfo of
     Just (TypeAlloc allocator n) -> do
-         wrapMaybe arg >>= bool
-            (do
-              name'' <- genConversion (prime name') $
-                        literal $ M $ allocator <>
-                                    " :: " <> typeShow (io ft)
-              line $ "memcpy " <> name'' <> " " <> name' <> " " <> tshow n
-              return name'')
-             -- The semantics of this case are somewhat undefined.
-            (notImplementedError "Nullable inout structs not supported")
+      when (transfer arg == TransferNothing && not (argCallerAllocates arg)) $
+        -- Although meaningful, this is almost always a bug in the
+        -- introspection data. Better not generating code than
+        -- generating code that crashes.
+        notImplementedError "‘transfer-ownership=none’ ‘caller-allocates=0’ inout structs not supported"
+
+      wrapMaybe arg >>= bool
+        (do
+            name'' <- genConversion (prime name') $
+                      literal $ M $ allocator <>
+                          " :: " <> typeShow (io ft)
+            line $ "memcpy " <> name'' <> " " <> name' <> " " <> tshow n
+            return name'')
+        -- The semantics of this case are somewhat undefined.
+        (notImplementedError "Nullable inout structs not supported")
     Nothing -> do
       if argCallerAllocates arg
       then return name'
